@@ -3,7 +3,7 @@
 #
 #  (C) Paul Evans, 2021 -- leonerd@leonerd.org.uk
 
-package XS::Parse::Infix 0.15;
+package XS::Parse::Infix 0.16;
 
 use v5.14;
 use warnings;
@@ -102,6 +102,9 @@ These tokens will all yield an info structure, with the following fields:
       void                     *hookdata;
    };
 
+If the operator name contains any non-ASCII characters they are presumed to be
+in UTF-8 encoding. This will matter for deparse purposes.
+
 =cut
 
 =head1 PARSE HOOKS
@@ -110,6 +113,8 @@ The C<XSParseInfixHooks> structure provides the following fields which are
 used at various stages of parsing.
 
    struct XSParseInfixHooks {
+      U16 flags; U8 lhs_flags;  /* currently ignored */
+      U8 rhs_flags;
       enum XSParseInfixClassification cls;
 
       const char *wrapper_func_name;
@@ -121,6 +126,34 @@ used at various stages of parsing.
       OP *(*ppaddr)(pTHX);
    };
 
+=head2 Flags
+
+The C<flags> and C<lhs_flags> fields are currently ignored. They are defined
+simply to reserve the space in case they are used in a later version. They
+should be set to zero.
+
+The C<rhs_flags> field gives details on how to parse and handle the right-hand
+side of the operator syntax. It should be set to one of the following constants:
+
+=over 4
+
+=item XPI_OPERAND_TERM (0)
+
+Default. The operand is a term expression.
+
+=item XPI_OPERAND_TERM_LIST
+
+The operand is a term expression. It will be foced into list context,
+preserving the C<OP_PUSHMARK> at the beginning. This means that the ppfunc for
+this infix operator will have to C<POPMARK> to find that.
+
+=item XPI_OPERAND_LIST
+
+The operand is a list expression. It will be forced into list context, the
+same as above.
+
+=back
+
 =head2 The Selection Stage
 
 The C<cls> field gives a "classification" of the operator, suggesting what
@@ -161,23 +194,33 @@ used instead to construct a new BINOP of the C<OP_CUSTOM> type.
 Additionally, if the C<wrapper_func_name> field is set to a string, this gives
 the (fully-qualified) name for a function to be generated as part of
 registering the operator. This newly-generated function will act as a wrapper
-for the operator, which is assumed to take two simple scalar arguments. The
-result of invoking the function on those arguments will be determined by using
-the operator code.
+for the operator.
+
+For operators whose RHS is a scalar, the wrapper function is assumed to take
+two simple scalar arguments. The result of invoking the function on those
+arguments will be determined by using the operator code.
 
    $result = $lhs OP $rhs;
 
    $result = WRAPPERFUNC( $lhs, $rhs );
 
-This creates a convenience for allowing perls that do not support
-C<PL_infix_plugin> to still have access to the operator.
+For operators whose RHS is a list, the wrapper function takes at least one
+argument, possibly more. The first argument is the scalar on the LHS, and the
+remaining arguments, however many there are, form the RHS:
 
-The wrapper function also includes a call-checker, which attempts to inline
-the operator directly into the callsite. Thus, in simple cases where the
-function is called directly in cases such as the following, no C<ENTERSUB>
-overhead will be incurred, and the generated optree will be identical to that
-which would have been generated by using infix operator syntax directly, even
-on versions of perl that do not have C<PL_infix_plugin>:
+   $result = $lhs OP @rhs;
+
+   $result = WRAPPERFUNC( $lhs, @rhs );
+
+This creates a convenience for accessing the operator from perls that do not
+support C<PL_infix_plugin>.
+
+In the case of scalar infix operators, the wrapper function also includes a
+call-checker which attempts to inline the operator directly into the callsite.
+Thus, in simple cases where the function is called directly on exactly two
+scalar arguments (such as in the following), no C<ENTERSUB> overhead will be
+incurred and the generated optree will be identical to that which would have
+been generated by using infix operator syntax directly:
 
    WRAPPERFUNC( $lhs, $rhs );
    WRAPPERFUNC( $lhs, CONSTANT );
@@ -195,6 +238,49 @@ forms:
 
 =cut
 
+=head1 DEPARSE
+
+This module operates with L<B::Deparse> in order to automatically provide
+deparse support for infix operators. Every infix operator that is implemented
+as a custom op (and thus has the C<ppaddr> hook field set) will have deparse
+logic added. This will allow it to deparse to either the named wrapper
+function, or to the infix operator syntax if on a C<PL_infix_plugin>-enabled
+perl and the appropriate lexical hint is enabled at the callsite.
+
+In order for this to work, it is important that your custom operator is I<not>
+registered as a custom op using the C<Perl_register_custom_op()> function.
+This registration will be performed by C<XS::Parse::Infix> itself at the time
+the infix operator is registered.
+
+=cut
+
+sub B::Deparse::_deparse_infix_wrapperfunc
+{
+   my ( $self, $wrapper_func_name, $op, $ctx ) = @_;
+
+   my $lhs = $op->first;
+   my $rhs = $op->last;
+
+   # Inspired by B::Deparse::pp_entersub
+
+   my $args = join ", ", map { $self->deparse( $_, 6 ) } ( $lhs, $rhs );
+
+   return "$wrapper_func_name($args)";
+}
+
+sub B::Deparse::_deparse_infix_named
+{
+   my ( $self, $opname, $op, $ctx ) = @_;
+
+   my $lhs = $op->first;
+   my $rhs = $op->last;
+
+   return join " ",
+      $self->deparse_binop_left( $op, $lhs, 14 ),
+      $opname,
+      $self->deparse_binop_right( $op, $rhs, 14 );
+}
+
 =head1 AUTHOR
 
 Paul Evans <leonerd@leonerd.org.uk>
